import { style } from 'typestyle';

export const historySideBarStyle = style({
  display: 'flex',
  flexDirection: 'column',

  minHeight: '400px',

  marginBlockStart: 0,
  marginBlockEnd: 0,
  paddingLeft: 0,

  overflowY: 'auto'
});
