.jp-git-PlainText-diff {
  width: 100%;
  max-height: calc(100% - 23px);
}

.jp-git-PlainText-diff .CodeMirror-merge,
.jp-git-PlainText-diff .CodeMirror-merge .CodeMirror,
.jp-git-PlainText-diff .CodeMirror-merge .CodeMirror-merge-pane {
  height: 100%;
}

.jp-git-PlainText-diff .CodeMirror-merge-gap {
  z-index: 2;
  display: inline-block;
  height: 100%;
  -moz-box-sizing: border-box;
  box-sizing: border-box;
  overflow: hidden;
  border-left: var(--codemirror-border);
  border-right: var(--codemirror-border);
  position: relative;
  color: var(--jp-ui-font-color2);
  background: var(--jp-layout-color2);
}

.jp-git-PlainText-diff .CodeMirror-merge-r-inserted,
.jp-git-PlainText-diff .CodeMirror-merge-l-inserted {
  background-position: bottom left;
  background-repeat: repeat-x;
}

.jp-git-PlainText-diff .CodeMirror-merge-r-deleted,
.jp-git-PlainText-diff .CodeMirror-merge-l-deleted {
  background-position: bottom left;
  background-repeat: repeat-x;
}

.jp-git-PlainText-diff .CodeMirror-merge-left .CodeMirror-merge-l-chunk {
  background-color: var(--jp-git-diff-deleted-color);
}

.jp-git-PlainText-diff .CodeMirror-merge-l-chunk {
  background-color: var(--jp-git-diff-added-color);
}

.CodeMirror-patchgutter {
  width: 0.7em;
}

.CodeMirror-patchgutter-insert::before {
  content: '+';
  color: #999;
  background-color: var(--jp-git-diff-added-color);
}

.CodeMirror-patchgutter-delete::before {
  content: '-';
  color: #999;
  background-color: var(--jp-git-diff-deleted-color);
}
