import { style } from 'typestyle';

export const fullscreenProgressClass = style({
  position: 'absolute',
  top: '50%',
  left: '50%',
  color: 'var(--jp-ui-inverse-font-color0)',
  textAlign: 'center'
});
