/* Import same style from nbdime and nbdime-jupyterlab (see index.ts) */
@import url('~nbdime/lib/common/collapsible.css');
@import url('~nbdime/lib/upstreaming/flexpanel.css');
@import url('~nbdime/lib/common/dragpanel.css');
@import url('~nbdime/lib/styles/variables.css');
@import url('~nbdime/lib/styles/common.css');
@import url('~nbdime/lib/styles/diff.css');
@import url('~nbdime/lib/styles/merge.css');
@import url('~nbdime-jupyterlab/style/index.css');

@import url('base.css');
