import { style } from 'typestyle';

export const gitWidgetStyle = style({
  display: 'flex',
  flexDirection: 'column',
  minWidth: '300px',
  color: 'var(--jp-ui-font-color1)',
  background: 'var(--jp-layout-color1)',
  fontSize: 'var(--jp-ui-font-size1)'
});
