(self.webpackChunk_jupyterlab_git=self.webpackChunk_jupyterlab_git||[]).push([[872],{3349:(t,e,r)=>{"use strict";function n(t){if(void 0===t)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}r.d(e,{Z:()=>n})},2122:(t,e,r)=>{"use strict";function n(){return(n=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var r=arguments[e];for(var n in r)Object.prototype.hasOwnProperty.call(r,n)&&(t[n]=r[n])}return t}).apply(this,arguments)}r.d(e,{Z:()=>n})},1788:(t,e,r)=>{"use strict";r.d(e,{Z:()=>o});var n=r(4665);function o(t,e){t.prototype=Object.create(e.prototype),t.prototype.constructor=t,(0,n.Z)(t,e)}},9756:(t,e,r)=>{"use strict";function n(t,e){if(null==t)return{};var r,n,o={},u=Object.keys(t);for(n=0;n<u.length;n++)r=u[n],e.indexOf(r)>=0||(o[r]=t[r]);return o}r.d(e,{Z:()=>n})},4665:(t,e,r)=>{"use strict";function n(t,e){return(n=Object.setPrototypeOf||function(t,e){return t.__proto__=e,t})(t,e)}r.d(e,{Z:()=>n})}}]);